/*
 * tkMacMDEF.r --
 *
 *	This file contains the actual MDEF. Since this is not likely to 
 *	change much, this seems the easiest method to use. The address
 *	of the routine descriptor is written into offset 0x24 hex, and
 *	then when the MDEF is called, the Mixed Mode Manager will take
 *	care of the setup.
 *
 *	This file also contains the icons 'SICN' used by the menu code
 *	in menu items.  
 *
 * Copyright (c) 1996-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: tkMacMDEF.r,v 1.2 1998/09/14 18:23:37 stanton Exp $
 */

#include <Types.r>

/*
 * This code was generated by a project file and will not need to be changed.
 * It is just a stub. The address of the real MDEF handler will be blasted
 * in.
 */

data 'MDEF' (591, preload) {
	$"600A 0000 4D44 4546 024F 0000 4EFA 0004"            /* `...MDEF.O..Nú.. */
	$"4E75 4E56 FFEE 48E7 1830 362E 0018 246E"            /* NuNVÿîHç.06...$n */
	$"0008 267C 4D44 4546 594F 2EB8 0B54 201F"            /* ..&|MDEFYO.¸.T . */
	$"2D40 FFEE 554F 3EB8 0A0A 301F 3D40 FFF2"            /* -@ÿîUO>¸..0.=@ÿò */
	$"554F 3EB8 0A0C 301F 3D40 FFF4 0C78 FFFF"            /* UO>¸..0.=@ÿô.xÿÿ */
	$"0B5C 6612 2078 0B5C 2050 2D68 0006 FFF6"            /* .\f. x.\ P-h..ÿö */
	$"2D68 000A FFFA 0C43 0001 6602 3812 3F03"            /* -h..ÿú.C..f.8.?. */
	$"2F2E 0014 2F2E 0010 2F2E 000C 2F0A 486E"            /* /.../.../.../.Hn */
	$"FFEE 4E93 2F2E FFEE 21DF 0B54 3F2E FFF2"            /* ÿîN“/.ÿî!ß.T?.ÿò */
	$"31DF 0A0A 3F2E FFF4 31DF 0A0C 0C43 0001"            /* 1ß..?.ÿô1ß...C.. */
	$"662A B852 6726 0C78 FFFF 0B5C 671E 2078"            /* f*¸Rg&.xÿÿ.\g. x */
	$"0B5C 2050 216E FFF6 0006 216E FFFA 000A"            /* .\ P!nÿö..!nÿú.. */
	$"21EE FFF6 09FA 21EE FFFA 09FE 4CDF 0C18"            /* !îÿöÆú!îÿúÆþLß.. */
	$"4E5E 205F 4FEF 0012 4ED0 846D 6169 6E00"            /* N^ _Oï..NÐ„main. */
	$"0000"                                               /* .. */
};
